"""World3 initialization."""

from whynot.simulators.world3.simulator import Config, Intervention, simulate, State
from whynot.simulators.world3.experiments import *
from whynot.simulators.world3.environments import *

SUPPORTS_CAUSAL_GRAPHS = False
