"""World2 initialization."""

from whynot.simulators.world2.simulator import (
    Config,
    Intervention,
    quality_of_life,
    simulate,
    State,
)
from whynot.simulators.world2.experiments import *

SUPPORTS_CAUSAL_GRAPHS = False
